from __future__ import annotations


class ModuleDocFragment(object):

    # Standard documentation fragment
    RETURN = r'''
y_notlast:
    description: A return from fragment
    type: str
    returned: it depends TM

z_last:
    description: A a return from fragment with merge.
    type: str
    returned: success
'''
